#!/usr/bin/env python
#
#        find-orphaned.py - A Gentoo Portage utility to find files
#                 not belonging to a installed package.
#
#  Copyright (C) 2010      Angelo Arrifano <miknix@gentoo.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  This is still a proof-of-concept, needs some love on exceptions and
#  command line arguments.
#  If you have comments, patches or beer; ping me at:
#  miknix@gmail.com or miknix@gentoo.org
#

import os
import sys

class Node(dict):
	def __init__(self, name):
		dict(self)
		self._name = name
		self._owner = None
	
	def add_child(self, name):
		if not self.has_key(name):
			node = Node(name)
			self[name] = node
			return node
		else:
			return self[name]

	def set_owner(self, owner):
		self._owner = owner

	def get_owner(self):
		return self._owner

class Cache:
	def __init__(self):
		self._root = Node('/')

	def add_pkg(self, pkg, files):
		for filepath in files:
			node = self.get_node(filepath)
			node.set_owner(pkg)
	
	def get_node(self, path):
		return self._get_node(self._root, path.split('/')[1:])

	def _get_node(self, node, path):
		# Holly functional programming, must love it
		if not path:
			return node
		return self._get_node(node.add_child(path[0]), path[1:])


##############################################################################
############################# Build cache ####################################
##############################################################################
def parse_line(line):
	# What if paths have spaces in CONTENTS?
	items = line.split(' ')
	if items[0] != 'obj' and items[0] != 'sym':
		return None
	return items[1]

def get_files(contentsfile):
	files = []
	with open(contentsfile, 'r') as f:
		for line in f:
			file = parse_line(line)
			if file:
				files.append(file)
	return files

cache = Cache()
portdb = '/var/db/pkg'
cats = os.listdir(portdb)
try:
	for cat in cats:
		catpath = os.path.join(portdb, cat)
		pkgs = os.listdir(catpath)
		for pkg in pkgs:
			print "* Caching %s/%s .." % (cat, pkg)
			pkgfiles = get_files(os.path.join(catpath, pkg, 'CONTENTS'))
			cache.add_pkg(pkg, pkgfiles)
except KeyboardInterrupt:
	print "Interrupted by user"
	sys.exit(1)

##############################################################################
########################## Filesystem scanner ################################
##############################################################################

def scan_dir(path):
	try:
		items = os.listdir(path)
	except OSError as (errno, strerror):
		if errno == 13: # Permission denied
			return
	for item in items:
		itempath = os.path.join(path, item)
		if os.path.isdir(itempath):
			scan_dir(itempath)
		else:
			node = cache.get_node(itempath)
			if not node.get_owner():
				print itempath

print "\n* And here is a list of orphaned files:"
try:
	for target in sys.argv[1:]:
		scan_dir(target)
except KeyboardInterrupt:
	print "Interrupted by user"
